"""
`epicsmacrolib help` will show detailed help information.
"""

import argparse

DESCRIPTION = __doc__


def build_arg_parser(argparser=None):
    if argparser is None:
        argparser = argparse.ArgumentParser()

    argparser.description = DESCRIPTION
    argparser.formatter_class = argparse.RawTextHelpFormatter

    argparser.add_argument(
        'argument_name',
        type=str,
        help='Get help on this.',
    )

    return argparser


def main(argument_name):
    print(f"This should show help for {argument_name!r}.")



import collections.abc
import contextlib
import dataclasses
from typing import Dict, Optional, Union

from _epicsmacrolib.macro import _MacroContext


@dataclasses.dataclass(frozen=True)
class MacroEntry:
    name: str
    rawval: str
    value: str
    type: str


class MacroContext(_MacroContext, collections.abc.MutableMapping):
    """
    A context for using EPICS macLib handle from Python.

    When other macro expansion tools just aren't enough, go with the one true
    macro expander - the one provided by EPICS - macLib.

    Parameters
    ----------
    use_environment : bool, optional
        Include environment variables when expanding macros.

    show_warnings : bool, optional
        Show warnings (see ``macSuppressWarning``).

    string_encoding : str, optional
        The default string encoding to use.  Defaults to latin-1, as these
        tools were written before utf-8 was really a thing.
    """

    #: Show warnings
    show_warnings: bool
    #: String encoding to use internally with macLib.
    string_encoding: str
    #: Include environment variables as macros.
    use_environment: bool

    def __init__(
        self,
        use_environment: bool = True,
        show_warnings: bool = False,
        string_encoding: str = "latin-1",
        macro_string: Optional[str] = None,
        macros: Optional[Dict[str, str]] = None,
    ):
        super().__init__(
            use_environment=use_environment,
            show_warnings=show_warnings,
            string_encoding=string_encoding,
        )

        if macros is not None:
            self.define(**macros)

        if macro_string is not None:
            self.define_from_string(macro_string)

    @contextlib.contextmanager
    def scoped(self, **macros: str):
        """A context manager to define macros (as kwargs) in a given scope."""
        self._push_scope()
        try:
            self.define(**macros)
            yield
        finally:
            self._pop_scope()

    def definitions_to_dict(
        self,
        defn: Union[str, bytes],
        string_encoding: str = ""
    ) -> Dict[str, str]:
        """
        Convert a definition string of the form ``A=value_a,B=value_a`` to a
        dictionary.

        Parameters
        ----------
        defn : Union[str, bytes]
            Definition string of the form ``A=value_a,B=value_a``.
        string_encoding : str, optional
            Encoding to use for the string, as macro lib works with bytes.
            Defaults to ``self.string_encoding``.

        Returns
        -------
        dict[str, str]
            Dictionary of name to value
        """
        return self._definitions_to_dict(
            defn,
            string_encoding=string_encoding or self.string_encoding,
        )

    def define_from_string(self, macro_string: str) -> Dict[str, str]:
        """Define macros with the standard VAR=VALUE syntax."""
        definitions = self.definitions_to_dict(macro_string)
        self.define(**definitions)
        return definitions

    def define(self, **macros: str) -> Dict[str, str]:
        """Use kwargs to define macros."""
        return super().define(**macros)

    def get_macro_details(self) -> Dict[str, MacroEntry]:
        """
        Get a dictionary of all MacroEntry items.

        This represents the internal state of the MAC_ENTRY nodes.

        Entry attributes include: name, rawval, value, type.
        """
        return super().get_macro_details()

    def get_macros(self) -> Dict[str, str]:
        """Get macros as a dictionary."""
        return dict(
            (macro.name, macro.value)
            for macro in self.get_macro_details().values()
        )

    def expand(
        self,
        value: str,
        *,
        empty_on_failure: bool = False,
        max_length: int = 1024,
    ) -> str:
        """
        Expand (interpolate) a string, replacing macros with values.

        By default, use the implicit buffer length of 1024 used in EPICS.

        Note that EPICS performs string interpolation line-by-line and not on a
        full-file basis.  See also ``expand_by_line``.

        Parameters
        ----------
        value : str
            The string to interpolate.
        empty_on_failure : bool, optional
            On failure (due to buffer size, a failed macExpandString result),
            return an empty string.
        max_length : int, optional
            Maximum length to use in macExpandString.

        Returns
        -------
        str
            The resulting interpolated string.
        """
        if max_length == 1024:
            return self._expand(value, empty_on_failure=empty_on_failure)
        return self._expand_with_length(
            value, empty_on_failure=empty_on_failure, max_length=max_length
        )

    def expand_by_line(self, contents: str, *, delimiter: str = "\n") -> str:
        """
        Expand a multi-line string, line-by-line.

        EPICS performs string interpolation line-by-line and not on a full-file
        basis.

        Parameters
        ----------
        contents : str
            The multi-line string.
        delimiter : str, optional
            The delimiter between lines.  Defaults to "\n".

        Returns
        -------
        str
            Interpolated multi-line string.
        """
        return delimiter.join(
            self.expand(line)
            for line in contents.splitlines()
        )


def macros_from_string(
    macro_string: str,
    use_environment: bool = False
) -> Dict[str, str]:
    """
    Get a macro dictionary from a macro string.

    Parameters
    ----------
    macro_string : str
        The macro string, in the format A=B,C=D,...

    use_environment : bool, optional
        Use environment variables as well.  Defaults to False.

    Returns
    -------
    macros : Dict[str, str]
        Macro key to value.
    """
    if not macro_string.strip():
        return {}
    macro_context = MacroContext(use_environment=use_environment)
    return macro_context.define_from_string(macro_string)


__all__ = ["MacroContext", "macros_from_string"]



import pytest

from .. import IocshRedirect, IocshSplit, split_iocsh_line


@pytest.mark.parametrize(
    "line, expected",
    [
        pytest.param(
            """dbLoadRecords(a,  "b", "c")""",
            ["dbLoadRecords", "a", "b", "c"],
            id="basic_paren"
        ),
        pytest.param(
            """dbLoadRecords a,  "b", "c\"""",
            ["dbLoadRecords", "a", "b", "c"],
            id="basic_no_paren"
        ),
        pytest.param(
            """< input_file""",
            IocshSplit(
                [],
                redirects={
                    0: IocshRedirect(fileno=0, name="input_file", mode="r"),
                },
                error=None,
            ),
            id="basic_input_redirect",
        ),
        pytest.param(
            """> output_file""",
            IocshSplit(
                [],
                redirects={
                    1: IocshRedirect(fileno=1, name="output_file", mode="w"),
                },
                error=None,
            ),
            id="basic_output_redirect",
        ),
        pytest.param(
            """< input_file > output_file""",
            IocshSplit(
                [],
                redirects={
                    0: IocshRedirect(fileno=0, name="input_file", mode="r"),
                    1: IocshRedirect(fileno=1, name="output_file", mode="w"),
                },
                error=None,
            ),
            id="input_output_redirect",
        ),
        pytest.param(
            """2> output_file""",
            IocshSplit(
                [],
                redirects={
                    2: IocshRedirect(fileno=2, name="output_file", mode="w"),
                },
                error=None,
            ),
            id="output_fd_num",
        ),
        pytest.param(
            """test > stdout 2> stderr 3> whoknows""",
            IocshSplit(
                ["test"],
                redirects={
                    1: IocshRedirect(fileno=1, name="stdout", mode="w"),
                    2: IocshRedirect(fileno=2, name="stderr", mode="w"),
                    3: IocshRedirect(fileno=3, name="whoknows", mode="w"),
                },
                error=None,
            ),
            id="output_fd_num_more",
        ),
    ]
)
def test_split_words(line, expected):
    if isinstance(expected, list):
        expected = IocshSplit(
            argv=expected,
            redirects={},
            error=None,
        )
    assert split_iocsh_line(line) == expected



"""
`epicsmacrolib expand` will interpolate text given macros.
"""

import argparse
import sys
from typing import List, Optional

from ..macro import MacroContext

DESCRIPTION = __doc__


def build_arg_parser(argparser=None):
    if argparser is None:
        argparser = argparse.ArgumentParser()

    argparser.description = DESCRIPTION
    argparser.formatter_class = argparse.RawTextHelpFormatter

    group = argparser.add_mutually_exclusive_group()
    group.add_argument(
        "-s", "--string",
        dest="text_string",
        required=False,
        type=str,
        help="Text string to expand, in place of --file",
    )

    group.add_argument(
        "-f", "--file",
        type=str,
        dest="filename",
        required=False,
        help="Filename to expand. Use - for standard input.",
    )

    argparser.add_argument(
        "-m", "--macro",
        type=str,
        dest="macros",
        action="append",
        help="Define one or more macros in the form A=B...",
    )

    argparser.add_argument(
        "--use-env",
        action="store_true",
        help="Include environment variables as macros",
    )

    argparser.add_argument(
        "-d", "--delimiter",
        default="\n",
        help="Multi-line delimiter.",
    )

    return argparser


def main(
    filename: Optional[str] = None,
    text_string: Optional[str] = None,
    macros: Optional[List[str]] = None,
    use_env: bool = False,
    delimiter: str = "\n",
):

    if filename == "-":
        text_string = sys.stdin.read()
    elif filename is not None:
        with open(filename, "rt") as fp:
            text_string = fp.read()

    if text_string is None:
        raise ValueError("Filename or text string required")

    ctx = MacroContext(use_environment=use_env)
    for macro in macros or []:
        ctx.define_from_string(macro)

    print(ctx.expand_by_line(text_string, delimiter=delimiter))



"""
`epicsmacrolib` is the top-level command for accessing various subcommands.

Try:

"""

import argparse
import importlib
import logging

import epicsmacrolib

DESCRIPTION = __doc__


MODULES = ("help", "expand")


def _try_import(module):
    relative_module = f'.{module}'
    return importlib.import_module(relative_module, 'epicsmacrolib.bin')


def _build_commands():
    global DESCRIPTION
    result = {}
    unavailable = []

    for module in sorted(MODULES):
        try:
            mod = _try_import(module)
        except Exception as ex:
            unavailable.append((module, ex))
        else:
            result[module] = (mod.build_arg_parser, mod.main)
            DESCRIPTION += f'\n    $ epicsmacrolib {module} --help'

    if unavailable:
        DESCRIPTION += '\n\n'

        for module, ex in unavailable:
            DESCRIPTION += (
                f'\nWARNING: "epicsmacrolib {module}" is unavailable due to:'
                f'\n\t{ex.__class__.__name__}: {ex}'
            )

    return result


COMMANDS = _build_commands()


def main():
    top_parser = argparse.ArgumentParser(
        prog='epicsmacrolib',
        description=DESCRIPTION,
        formatter_class=argparse.RawTextHelpFormatter
    )

    top_parser.add_argument(
        '--version', '-V',
        action='version',
        version=epicsmacrolib.__version__,
        help="Show the epicsmacrolib version number and exit."
    )

    top_parser.add_argument(
        '--log', '-l', dest='log_level',
        default='INFO',
        type=str,
        help='Python logging level (e.g. DEBUG, INFO, WARNING)'
    )

    subparsers = top_parser.add_subparsers(help='Possible subcommands')
    for command_name, (build_func, main) in COMMANDS.items():
        sub = subparsers.add_parser(command_name)
        build_func(sub)
        sub.set_defaults(func=main)

    args = top_parser.parse_args()
    kwargs = vars(args)
    log_level = kwargs.pop('log_level')

    logger = logging.getLogger('epicsmacrolib')
    logger.setLevel(log_level)
    logging.basicConfig()

    if hasattr(args, 'func'):
        func = kwargs.pop('func')
        logger.debug('%s(**%r)', func.__name__, kwargs)
        func(**kwargs)
    else:
        top_parser.print_help()


if __name__ == '__main__':
    main()



import os
import pathlib
import sys
from os import path

from setuptools import Extension, find_packages, setup

import versioneer

min_version = (3, 7)

if sys.version_info < min_version:
    error = """
epicsmacrolib does not support Python {0}.{1}.
Python {2}.{3} and above is required. Check your Python version like so:

python3 --version

This may be due to an out-of-date pip. Make sure you have pip >= 9.0.1.
Upgrade pip like so:

pip install --upgrade pip
""".format(
        *sys.version_info[:2], *min_version
    )
    sys.exit(error)


if sys.platform == "darwin":
    # Required for building on macOS - an appropriate deployment target:
    os.environ.setdefault("MACOSX_DEPLOYMENT_TARGET", "10.9")

here = path.abspath(path.dirname(__file__))

with open(path.join(here, "README.rst"), encoding="utf-8") as readme_file:
    readme = readme_file.read()

with open(path.join(here, "requirements.txt")) as requirements_file:
    # Parse requirements.txt, ignoring any commented-out lines.
    requirements = [
        line
        for line in requirements_file.read().splitlines()
        if not line.startswith("#")
    ]


git_requirements = [r for r in requirements if r.startswith("git+")]
if git_requirements:
    print("User must install the following packages manually:")
    print()
    print("\n".join(f"* {r}" for r in git_requirements))
    print()


# https://cython.readthedocs.io/en/latest/src/userguide/source_files_and_compilation.html#distributing-cython-modules
def no_cythonize(extensions, **_ignore):
    for extension in extensions:
        sources = []
        for sfile in extension.sources:
            path, ext = os.path.splitext(sfile)
            if ext in (".pyx", ".py"):
                if extension.language == "c++":
                    ext = ".cpp"
                else:
                    ext = ".c"
                sfile = f"{path}{ext}"
            sources.append(sfile)
        extension.sources[:] = sources
    return extensions


def get_extensions():
    """Get the list of Cython extensions to add."""
    SRC_DIR = pathlib.Path("src")
    ext_options = dict(
        include_dirs=["./src"],
        libraries=[],
        library_dirs=["./src"],
        language="c++",
    )

    macro_lib_sources = list(str(path) for path in SRC_DIR.glob("*.c"))
    extensions = [
        Extension("_epicsmacrolib.iocsh", ["epicsmacrolib/iocsh.pyx"], **ext_options),
        Extension(
            "_epicsmacrolib.macro",
            ["epicsmacrolib/macro.pyx"] + macro_lib_sources,
            **ext_options,
        ),
    ]

    should_cythonize = bool(int(os.getenv("CYTHONIZE", "0")))
    if should_cythonize:
        from Cython.Build import cythonize
        compiler_directives = {"language_level": 3, "embedsignature": True}
        return cythonize(extensions, compiler_directives=compiler_directives)
    return no_cythonize(extensions)


with open("requirements.txt") as fp:
    install_requires = [
        line for line in fp.read().splitlines() if line and not line.startswith("#")
    ]

with open("README.rst", encoding="utf-8") as fp:
    readme = fp.read()


setup(
    name="epicsmacrolib",
    version=versioneer.get_version(),
    cmdclass=versioneer.get_cmdclass(),
    license="BSD",
    author="SLAC National Accelerator Laboratory",
    packages=find_packages(exclude=["docs", "tests"]),
    description="epics-base compliant macro tools",
    long_description=readme,
    long_description_content_type="text/x-rst",
    url="https://github.com/pcdshub/epicsmacrolib",  # noqa
    entry_points={
        "console_scripts": [
            "epicsmacrolib=epicsmacrolib.bin.main:main",
        ],
    },
    include_package_data=True,
    package_data={
        "epicsmacrolib": [
            # When adding files here, remember to update MANIFEST.in as well,
            # or else they will not be included in the distribution on PyPI!
            # 'path/to/data_file',
        ]
    },
    install_requires=requirements,
    classifiers=[
        "Development Status :: 2 - Pre-Alpha",
        "Natural Language :: English",
        "Programming Language :: Python :: 3",
    ],
    ext_modules=get_extensions(),
)



from dataclasses import dataclass
from typing import Dict, List, Optional, TypedDict, cast

from _epicsmacrolib.iocsh import split_iocsh_line as _split_iocsh_line


@dataclass
class IocshRedirect:
    fileno: int
    name: str
    mode: str


@dataclass
class IocshSplit:
    argv: List[str]
    redirects: Optional[Dict[int, IocshRedirect]] = None
    error: Optional[str] = None


class _IocshRedirect(TypedDict):
    fileno: int
    name: str
    mode: str


def split_iocsh_line(
    input_line: str,
    string_encoding: str = 'latin-1',
    ifs: bytes = b" \t(),\r",
    num_redirects: int = 5,
) -> IocshSplit:
    """
    Split ``input_line`` into words, according to how the IOC shell would.

    Note that this is almost a direct conversion of the original C code, making
    an attempt to avoid introducing inconsistencies between this implementation
    and the original.

    Parameters
    ----------
    input_line : str
        The line to split.

    Returns
    -------
    info : IocshSplit
    """
    res = _split_iocsh_line(
        input_line,
        string_encoding=string_encoding,
        ifs=ifs,
        num_redirects=num_redirects,
    )
    split = IocshSplit(argv=res["argv"], redirects=None, error=res["error"])

    redirects = res["redirects"]
    if redirects is not None:
        redirects = cast(Dict[int, _IocshRedirect], redirects)
        split.redirects = {
            idx: IocshRedirect(**redirect)
            for idx, redirect in redirects.items()
        }
    return split



from . import _version

__version__ = _version.get_versions()['version']

from .iocsh import IocshRedirect as IocshRedirect  # noqa
from .iocsh import IocshSplit as IocshSplit  # noqa
from .iocsh import split_iocsh_line as split_iocsh_line
from .macro import MacroContext as MacroContext
from .macro import macros_from_string as macros_from_string

__all__ = ["MacroContext", "macros_from_string", "split_iocsh_line"]



import dataclasses
import os

import pytest

from .. import MacroContext


def test_env():
    os.environ["C"] = "3"
    os.environ["D"] = "4"
    os.environ["E"] = "5"

    ctx = MacroContext(use_environment=True)
    ctx.define_from_string("A=1,B=2,E=15")

    assert ctx["A"] == "1"
    assert ctx["B"] == "2"
    assert ctx["C"] == "3"
    assert ctx["D"] == "4"
    assert ctx["E"] == "15"

    assert set(list(ctx.items())[-6:]) == {
        ("A", "1"),
        ("B", "2"),
        ("C", "3"),
        ("D", "4"),
        ("E", "15"),
    }
    with pytest.raises(KeyError):
        ctx["F"]

    with ctx.scoped(**ctx.definitions_to_dict("A=11,C=13,D=14,G=7")):
        assert ctx["A"] == "11"
        assert ctx["B"] == "2"
        assert ctx["C"] == "13"
        assert ctx["D"] == "14"
        assert ctx["E"] == "15"
        with pytest.raises(KeyError):
            ctx["F"]
        assert ctx["G"] == "7"

        # implicit when called through in iocshBody
        os.environ["D"] = "24"
        ctx["D"] = "24"
        os.environ["F"] = "6"
        ctx["F"] = "6"
        os.environ["G"] = "17"
        ctx["G"] = "17"

        assert ctx["D"] == "24"
        assert ctx["F"] == "6"
        assert ctx["G"] == "17"

    assert ctx["A"] == "1"
    assert ctx["B"] == "2"
    assert ctx["C"] == "3"
    assert ctx["D"] == "24"
    assert ctx["E"] == "15"
    assert ctx["F"] == "6"
    assert ctx["G"] == "17"

    with ctx.scoped(**ctx.definitions_to_dict("D=34,G=27")):
        assert ctx["D"] == "34"
        assert ctx["G"] == "27"

    assert ctx["D"] == "24"


def test_empty():
    ctx = MacroContext(use_environment=False)
    with pytest.raises(KeyError):
        ctx["A"]
    with pytest.raises(KeyError):
        ctx["HOSTNAME"]


def test_misc_with_env():
    def check(macro_string, macros, expected):
        ctx = MacroContext(use_environment=True)
        ctx.define_from_string(macros)
        expanded = ctx.expand(macro_string, empty_on_failure=True)
        print(f"{macro_string!r} expanded to {expanded!r} with {macros!r}")
        assert expanded == expected

    os.environ.pop("FOO", None)
    check("FOO", "", "FOO")

    # NOTE: a bit of an API break here, we always want a string from expand,
    # but the original would return NULL (could be Python None)
    check("${FOO}", "", "")
    check("${FOO,BAR}", "", "")
    check("${FOO,BAR=baz}", "", "")
    check("${FOO,BAR=$(FOO)}", "", "")
    check("${FOO,FOO}", "", "")
    check("${FOO,FOO=$(FOO)}", "", "")
    check("${FOO,BAR=baz,FUM}", "", "")

    check("${=}", "", "")
    check("x${=}y", "", "xy")

    check("${,=}", "", "")
    check("x${,=}y", "", "xy")

    check("${FOO=}", "", "")
    check("x${FOO=}y", "", "xy")

    check("${FOO=,}", "", "")
    check("x${FOO=,}y", "", "xy")

    check("${FOO,FOO=}", "", "")
    check("x${FOO,FOO=}y", "", "xy")

    check("${FOO=,BAR}", "", "")
    check("x${FOO=,BAR}y", "", "xy")

    check("${FOO=$(BAR=)}", "", "")
    check("x${FOO=$(BAR=)}y", "", "xy")

    check("${FOO=,BAR=baz}", "", "")
    check("x${FOO=,BAR=baz}y", "", "xy")

    check("${FOO=$(BAR),BAR=}", "", "")
    check("x${FOO=$(BAR),BAR=}y", "", "xy")

    check("${=BAR}", "", "BAR")
    check("x${=BAR}y", "", "xBARy")

    check("${FOO=BAR}", "", "BAR")
    check("x${FOO=BAR}y", "", "xBARy")

    os.environ["FOO"] = "BLETCH"
    check("${FOO}", "", "BLETCH")
    check("${FOO,FOO}", "", "BLETCH")
    check("x${FOO}y", "", "xBLETCHy")
    check("x${FOO}y${FOO}z", "", "xBLETCHyBLETCHz")
    check("${FOO=BAR}", "", "BLETCH")
    check("x${FOO=BAR}y", "", "xBLETCHy")
    check("${FOO=${BAZ}}", "", "BLETCH")
    check("${FOO=${BAZ},BAR=$(BAZ)}", "", "BLETCH")
    check("x${FOO=${BAZ}}y", "", "xBLETCHy")
    check("x${FOO=${BAZ},BAR=$(BAZ)}y", "", "xBLETCHy")
    check("${BAR=${FOO}}", "", "BLETCH")
    check("x${BAR=${FOO}}y", "", "xBLETCHy")
    check("w${BAR=x${FOO}y}z", "", "wxBLETCHyz")

    check("${FOO,FOO=BAR}", "", "BAR")
    check("x${FOO,FOO=BAR}y", "", "xBARy")
    check("${BAR,BAR=$(FOO)}", "", "BLETCH")
    check("x${BAR,BAR=$(FOO)}y", "", "xBLETCHy")
    check("${BAR,BAR=$($(FOO)),BLETCH=GRIBBLE}", "", "GRIBBLE")
    check("x${BAR,BAR=$($(FOO)),BLETCH=GRIBBLE}y", "", "xGRIBBLEy")
    check("${$(BAR,BAR=$(FOO)),BLETCH=GRIBBLE}", "", "GRIBBLE")
    check("x${$(BAR,BAR=$(FOO)),BLETCH=GRIBBLE}y", "", "xGRIBBLEy")

    check("${FOO}/${BAR}", "BAR=GLEEP", "BLETCH/GLEEP")
    check("x${FOO}/${BAR}y", "BAR=GLEEP", "xBLETCH/GLEEPy")
    check("${FOO,BAR}/${BAR}", "BAR=GLEEP", "BLETCH/GLEEP")
    check("${FOO,BAR=x}/${BAR}", "BAR=GLEEP", "BLETCH/GLEEP")
    check("${BAZ=BLETCH,BAR}/${BAR}", "BAR=GLEEP", "BLETCH/GLEEP")
    check("${BAZ=BLETCH,BAR=x}/${BAR}", "BAR=GLEEP", "BLETCH/GLEEP")

    check("${${FOO}}", "BAR=GLEEP,BLETCH=BAR", "BAR")
    check("x${${FOO}}y", "BAR=GLEEP,BLETCH=BAR", "xBARy")
    check("${${FOO}=GRIBBLE}", "BAR=GLEEP,BLETCH=BAR", "BAR")
    check("x${${FOO}=GRIBBLE}y", "BAR=GLEEP,BLETCH=BAR", "xBARy")

    check("${${FOO}}", "BAR=GLEEP,BLETCH=${BAR}", "GLEEP")

    os.environ["FOO"] = "${BAR}"
    check("${FOO}", "BAR=GLEEP,BLETCH=${BAR}", "GLEEP")

    # NOTE: None case
    check("${FOO}", "BAR=${BAZ},BLETCH=${BAR}", "")

    check("${FOO}", "BAR=${BAZ=GRIBBLE},BLETCH=${BAR}", "GRIBBLE")

    check("${FOO}", "BAR=${STR1},BLETCH=${BAR},STR1=VAL1,STR2=VAL2", "VAL1")

    check("${FOO}", "BAR=${STR2},BLETCH=${BAR},STR1=VAL1,STR2=VAL2", "VAL2")

    # NOTE: None cases
    check("${FOO}", "BAR=${FOO},BLETCH=${BAR},STR1=VAL1,STR2=VAL2", "")
    check("${FOO,FOO=$(FOO)}", "BAR=${FOO},BLETCH=${BAR},STR1=VAL1,STR2=VAL2", "")
    check("${FOO=$(FOO)}", "BAR=${FOO},BLETCH=${BAR},STR1=VAL1,STR2=VAL2", "")
    check("${FOO=$(BAR),BAR=$(FOO)}", "BAR=${FOO},BLETCH=${BAR},STR1=VAL1,STR2=VAL2", "")
    os.environ.pop("FOO", None)


def test_not_empty_on_failure():
    def check(macro_string, macros, expected):
        ctx = MacroContext(use_environment=True)
        ctx.define_from_string(macros)
        expanded = ctx.expand(macro_string, empty_on_failure=False)
        print(f"{macro_string!r} expanded to {expanded!r} with {macros!r}")
        if expected:
            assert expanded == expected

    os.environ.pop("FOO", None)
    check("${FOO}", "", "$(FOO)")
    check("${FOO,BAR}", "", "$(FOO)")
    check("${FOO,BAR=baz}", "", "$(FOO)")
    check("${FOO,BAR=$(FOO)}", "", "$(FOO)")
    check("${FOO,FOO}", "", "$(FOO)")
    check("${FOO,FOO=$(FOO)}", "", "$(FOO)")
    check("${FOO,BAR=baz,FUM}", "", "$(FOO)")

    os.environ["FOO"] = "${BAR}"
    check("${FOO}", "BAR=${BAZ},BLETCH=${BAR}", "$(BAZ)")
    check("${FOO}", "BAR=${FOO},BLETCH=${BAR},STR1=VAL1,STR2=VAL2", "$(BAR)")
    check("${FOO,FOO=$(FOO)}", "BAR=${FOO},BLETCH=${BAR},STR1=VAL1,STR2=VAL2", "$(FOO)")
    check("${FOO=$(FOO)}", "BAR=${FOO},BLETCH=${BAR},STR1=VAL1,STR2=VAL2", "$(BAR)")
    check("${FOO=$(BAR),BAR=$(FOO)}", "BAR=${FOO},BLETCH=${BAR},STR1=VAL1,STR2=VAL2", "$(BAR)")

    os.environ.pop("FOO", None)


def test_not_empty_on_failure_show_warnings():
    def check(macro_string, macros, expected):
        ctx = MacroContext(use_environment=True, show_warnings=True)
        ctx.define_from_string(macros)
        expanded = ctx.expand(macro_string, empty_on_failure=False)
        print(f"{macro_string!r} expanded to {expanded!r} with {macros!r}")
        if expected:
            assert expanded == expected

    os.environ.pop("FOO", None)
    check("${FOO}", "", "$(FOO,undefined)")
    check("${FOO,BAR}", "", "$(FOO,undefined)")
    check("${FOO,BAR=baz}", "", "$(FOO,undefined)")
    check("${FOO,BAR=$(FOO)}", "", "$(FOO,undefined)")
    check("${FOO,FOO}", "", "$(FOO,undefined)")
    check("${FOO,FOO=$(FOO)}", "", "$(FOO,recursive)")
    check("${FOO,BAR=baz,FUM}", "", "$(FOO,undefined)")

    os.environ["FOO"] = "${BAR}"
    check("${FOO}", "BAR=${BAZ},BLETCH=${BAR}", "$(BAZ,undefined)")
    check("${FOO}", "BAR=${FOO},BLETCH=${BAR},STR1=VAL1,STR2=VAL2", "$(BAR,recursive)")
    check(
        "${FOO,FOO=$(FOO)}",
        "BAR=${FOO},BLETCH=${BAR},STR1=VAL1,STR2=VAL2",
        "$(FOO,recursive)",
    )
    check(
        "${FOO=$(FOO)}",
        "BAR=${FOO},BLETCH=${BAR},STR1=VAL1,STR2=VAL2",
        "$(BAR,recursive)",
    )
    check(
        "${FOO=$(BAR),BAR=$(FOO)}",
        "BAR=${FOO},BLETCH=${BAR},STR1=VAL1,STR2=VAL2",
        "$(BAR,recursive)",
    )

    os.environ.pop("FOO", None)


def test_python_niceties():
    ctx = MacroContext(use_environment=False)
    ctx["B"] = "5"
    with pytest.raises(KeyError):
        ctx["A"]
    assert ctx["B"] == "5"

    ctx.define(C="6")
    assert ctx["C"] == "6"

    assert ctx.items()
    assert set(list(ctx.items())) == {
        ("B", "5"),
        ("C", "6"),
    }

    assert set(ctx) == {"B", "C"}


def test_readme_example():
    from epicsmacrolib import MacroContext

    ctx = MacroContext(use_environment=True)
    ctx.define(TEST="A")
    print(ctx.expand("TEST=$(TEST) SHELL=$(SHELL)"))
    # TEST=A SHELL=/bin/bash

    ctx = MacroContext(use_environment=False)
    ctx.define_from_string("A=5,B=6")
    ctx.define(C="7")
    print(ctx.expand("$(A) $(B) ${C} ${D=5} ${E}"))
    # -> 5 6 7 5 $(E)

    ctx = MacroContext(use_environment=False, show_warnings=True)
    ctx.define_from_string("A=5,B=6")
    ctx.define(C="7")
    print(ctx.expand("$(A) $(B) ${C} ${D=5} ${E}"))
    # -> 5 6 7 5 $(E,undefined)

    ctx.define_from_string("A=5,B=$(B)")
    print(ctx.expand("$(A) $(B)"))
    # -> 5 $(B,recursive)

    with ctx.scoped(A="10", B="0"):
        print(ctx.expand("$(A)"))
        # -> 10
        with ctx.scoped(A="0"):
            print(ctx.expand("$(A)"))
            # -> 0
        print(ctx.expand("$(A)"))
        # -> 10

    for key, info in ctx.get_macro_details().items():
        print(key, info, dataclasses.asdict(info))
